@@ -109,16 +109,15 @@ module Test
           super
         end
 
-        @@added_methods = {}
+        @@added_method_names = {}
         def method_added(name) # :nodoc:
           super
-          _added_methods = added_methods
+          added_method_names = (@@added_method_names[self] ||= {})
           stringified_name = name.to_s
-          if _added_methods.include?(stringified_name)
+          if added_method_names.key?(stringified_name)
             attribute(:redefined, {:backtrace => caller}, {}, stringified_name)
           end
-          _attributes = attributes_table[stringified_name] || {}
-          source_location = _attributes[:source_location]
+          source_location = find_attribute(stringified_name, :source_location)
           if source_location
             path, line = source_location
           else
@@ -130,11 +129,11 @@ module Test
             :path => path,
             :line => line,
           }
-          _added_methods << stringified_name
+          added_method_names[stringified_name] = true
         end
 
-        def added_methods # :nodoc:
-          @@added_methods[self] ||= []
+        def added_method_names # :nodoc:
+          (@@added_method_names[self] ||= {}).keys
         end
 
         # Rolls up all of the test* methods in the fixture into
@@ -321,7 +320,7 @@ module Test
         #     end
         #   end
         #
-        # The diffrence of them are the following:
+        # The difference of them are the following:
         #
         # * Test case created by {sub_test_case} is an anonymous class.
         #   So you can't refer the test case by name.
@@ -332,14 +331,14 @@ module Test
         #
         # @param name [String] The name of newly created sub test case.
         # @yield
-        #   The block is evaludated under the newly created sub test
+        #   The block is evaluated under the newly created sub test
         #   case class context.
         # @return [Test::Unit::TestCase] Created sub test case class.
         def sub_test_case(name, &block)
           parent_test_case = self
           sub_test_case = Class.new(self) do
             singleton_class = class << self; self; end
-            singleton_class.send(:define_method, :name) do
+            singleton_class.__send__(:define_method, :name) do
               [parent_test_case.name, name].compact.join("::")
             end
           end
@@ -347,7 +346,7 @@ module Test
           sub_test_case
         end
 
-        # Checkes whether a test that is mathched the query is
+        # Checks whether a test that is matched the query is
         # defined.
         #
         # @option query [String] :path (nil)
@@ -361,17 +360,16 @@ module Test
           query_line = query[:line]
           query_method_name = query[:method_name]
 
-          available_locations = method_locations
-          if query_path
-            available_locations = available_locations.find_all do |location|
-              location[:path].end_with?(query_path)
-            end
-          end
+          available_locations = target_method_locations(query_path)
           if query_line
-            available_location = available_locations.reverse.find do |location|
+            available_locations = available_locations.sort_by do |location|
+              -location[:line]
+            end
+            available_location = available_locations.find do |location|
               query_line >= location[:line]
             end
             return false if available_location.nil?
+            return false if available_location[:test_case] != self
             available_locations = [available_location]
           end
           if query_method_name
@@ -392,6 +390,28 @@ module Test
         def method_locations
           @@method_locations[self] ||= []
         end
+
+        # @private
+        def target_method_locations(path)
+          if path.nil?
+            self_location = method_locations.first
+            path = self_location[:path] if self_location
+          end
+          return [] if path.nil?
+
+          target_locations = []
+          @@method_locations.each do |test_case, locations|
+            locations.each do |location|
+              absolete_path = File.expand_path(path)
+              location_path = location[:path]
+              location_basename = File.basename(location_path)
+              if location_path == absolete_path or location_basename == path
+                target_locations << location.merge(:test_case => test_case)
+              end
+            end
+          end
+          target_locations
+        end
       end
 
       attr_reader :method_name
@@ -704,7 +724,7 @@ module Test
           if handler.respond_to?(:call)
             handled = handler.call(self, exception)
           else
-            handled = send(handler, exception)
+            handled = __send__(handler, exception)
           end
           return true if handled
         end
